////////////////////////////////////////////////////////////////////////////
//   Copyright 2013  Teradici Corporation
//
//   This program is free software; you can redistribute it and/or modify it
//   under the terms and conditions of the GNU General Public License,
//   version 2, as published by the Free Software Foundation.
//
////////////////////////////////////////////////////////////////////////////
/// \file
/// \brief
///     GFUN Module Internal
///
///     This file defines the internal GFUN module function protoypes and
///     definitions.
///
////////////////////////////////////////////////////////////////////////////

#pragma once

// ---- Include Files -------------------------------------------------------

#include "tera_gfun.h"
#include "tera_mem.h"

// ---- Constants and Types -------------------------------------------------

// Number of DMA channels per GFUN (NOTE: Only supporting one channel for now)
#define GFUN_NUM_DMA_CHAN   1

// Length of ingress/egress descriptor queues
#define INGRESS_DESC_Q_LEN  20
#define EGRESS_DESC_Q_LEN   20

// --------------------------------------------------------------------------
//  sGFUN_DMA_DESC
///     GFUN DMA descriptor
//
//  Description:
///     This structure defines a GFUN DMA descriptor. This structure MUST match
///     the definition in the GFUN specification.
///
typedef struct sGFUN_DMA_DESC_STRUCT
{
    UINT64              buf_addr;       ///< 64 bit buffer address
    UINT16              buf_len;        ///< Total buffer length (lower 14 bits)
    UINT16              dma_disable;    ///< DMA disable flag (bit 0)
    UINT32              reserved;
    UINT16              len;            ///< Length of valid data in buffer
    UINT16              flags;          ///< Flags {0: eop; 15: valid}
    sTERA_GFUN_PKT_CTRL pkt_ctrl;       ///< Packet control (56 valid bits + padding)
    UINT32              pkt_err;        ///< Packet error (lower 7 bits)
} sGFUN_DMA_DESC;

#define GFUN_DMA_DESC_BUF_LEN_MASK      (0x3FFF)
#define GFUN_DMA_DESC_DMA_DISABLE_MASK  (0x0001)
#define GFUN_DMA_DESC_LEN_MASK          (0x3FFF)
#define GFUN_DMA_DESC_FLAGS_EOP_MASK    (0x0001)
#define GFUN_DMA_DESC_FLAGS_VALID_MASK  (0x8000)
#define GFUN_DMA_DESC_PKT_ERR_MASK      (0x007F)

// --------------------------------------------------------------------------
//  sTERA_GFUN_CBLK
///     GFUN module control block
//
//  Description:
///     This structure defines the GFUN module control block. The control block
///     contains the state flags, memory-mapped registers pointers, and also
///     pointers to the ingress and egress descriptor queues for all the GFUN
///     channels.
///
typedef struct sTERA_GFUN_CBLK_STRUCT
{
    struct pci_dev *dev;        ///< Device object
    UINT8     initialized;      ///< Initialization flag
    UINT8     started;          ///< Started flag

    UINT32    gfun_cap;         ///< GFUN capabilities

    UINT32    mtu_size[GFUN_NUM_DMA_CHAN];  ///< GFUN DMA channel size

    BYTE     *reg_base;         ///< Memory-mapped registers base address
    UINT32    reg_len;          ///< Memory-mapped registers length
    eTERA_PCOIP_DEVICE_TYPE dev_type; ///< PCoIP device type

    // Ingress descriptor queue and buffers
    sTERA_MEM_DMA *ingress_desc_q[GFUN_NUM_DMA_CHAN];
    sTERA_MEM_BUF *ingress_desc_q_buf[GFUN_NUM_DMA_CHAN][INGRESS_DESC_Q_LEN];

    // Egress descriptor queue and buffers
    sTERA_MEM_DMA *egress_desc_q[GFUN_NUM_DMA_CHAN];
    sTERA_MEM_BUF *egress_desc_q_buf[GFUN_NUM_DMA_CHAN][EGRESS_DESC_Q_LEN];

    /* for ioport free */
    int bars;
} sTERA_GFUN_CBLK;

// ---- Variable Externs ----------------------------------------------------

// ---- Function Prototypes -------------------------------------------------

