////////////////////////////////////////////////////////////////////////////
//   Copyright 2013  Teradici Corporation
//
//   This program is free software; you can redistribute it and/or modify it
//   under the terms and conditions of the GNU General Public License,
//   version 2, as published by the Free Software Foundation.
//
////////////////////////////////////////////////////////////////////////////
/// \file
/// \brief
///     Device Access Class Implementation
///
///     This file implements the device access class. This class represents
///     a device in kernel space that has a device interface. The purpose of
///     this class is to simplify the opening of the device, which is a
///     tedious task of enumerating through the devices in the system based
///     on a class GUID. This class also provides synchronous ioctls with a
///     timeout, something that is not native to Windows.
///
////////////////////////////////////////////////////////////////////////////

// ---- Include Files -------------------------------------------------------


#include <sys/ioctl.h>
#include <signal.h>
#include <fcntl.h>
#include <unistd.h>

#include "tera_errors.h"
#include "tera_log.h"
#include "tera_dev.h"

// ---- Global Variables ----------------------------------------------------

// ---- Private Constants and Types -----------------------------------------

// ---- Local Variables -----------------------------------------------------

// ---- Local Function Prototypes -------------------------------------------

// --------------------------------------------------------------------------
//  cTERA_DEV::cTERA_DEV
///     Constructor
//
//  Description:
///     This function is the class contructor. This function initializes the
///     member data variables. If the device path is provided, then this
///     function opens a handle to the device.
///
/// \return
///     None
///
cTERA_DEV::cTERA_DEV(
    const char *dev_path /* NULL */
    )
{
    INT32   ret;

    mTERA_LOG_DBG("start");

    // Initialize the member data variables
    m_initialized = false;
    m_ready = false;
    m_dev_fd = -1;

    if (dev_path != NULL)
    {
        // Initialize this object
        ret = init(dev_path);
        if (TERA_SUCCESS != ret)
        {
            mTERA_LOG_DBG("init() failed: returned %d", ret);
        }
    }

    mTERA_LOG_DBG("end");
} // cTERA_DEV::cTERA_DEV()

// --------------------------------------------------------------------------
//  cTERA_DEV::cTERA_DEV
///     Destructor
//
//  Description:
///     This function is the class destructor. This function cleans the
///     member data variables.
///
/// \return
///     None
///
cTERA_DEV::~cTERA_DEV(void)
{
    mTERA_LOG_DBG("start");

    if (m_dev_fd != -1)
    {
        close(m_dev_fd);
    }

    mTERA_LOG_DBG("end");
} // cTERA_DEV::~cTERA_DEV()

// --------------------------------------------------------------------------
//  cTERA_DEV::is_ready
///     Return if I/O operations with the device is available
//
//  Description:
///     This functions returns if I/O operations with the device is available.
///     If this object has not been initialized, or if an ioctl failed with
///     ERROR_NOT_READY, then this function returns 0, else 1.
///
/// \return
///     1,
///     0
///
INT cTERA_DEV::is_ready(void) const
{
    mTERA_LOG_DBG("m_initialized = %d, m_ready = %d", m_initialized, m_ready);

    // Being ready depends on both initialized and ready flags
    return (m_initialized && m_ready);
} // cTERA_DEV::is_ready()

// --------------------------------------------------------------------------
//  cTERA_DEV::init
///     Initialize this object
//
//  Description:
///     This function initializes this object by getting the path to the
///     device interface with the user specified GUID. This function must be
///     called before any I/O operations can be done on the device.
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_INVALID_ARG,
///     TERA_ERR_NOT_FOUND
///
INT32 cTERA_DEV::init(
    const char *dev_path    ///< Device pathname
    )
{
    INT32 ret;

    mTERA_LOG_DBG("start");

    if (dev_path == NULL)
    {
        ret = TERA_ERR_INVALID_ARG;
    }
    else if (m_initialized)
    {
        ret = TERA_SUCCESS;
    }
    else
    {
        // Open a file descriptor to the device
        m_dev_fd = open(dev_path, O_RDWR);
        if (m_dev_fd < 0)
        {
            ret = TERA_ERR_NOT_FOUND;
        }
        else
        {
            m_initialized = true;
            m_ready = true;

            ret = TERA_SUCCESS;
        }
    }

    mTERA_LOG_DBG("end");

    return ret;
} // cTERA_DEV::init()

// --------------------------------------------------------------------------
//  cTERA_DEV::synch_ioctl
///     Send a synchronous ioctl to the device
//
//  Description:
///     This function sends a synchronous ioctl to the device. The device must
///     be initialized before any I/O operations can be done. This function
///     uses the device path determined in the init function, and creates a
///     handle to the interface. This function will then send the ioctl and
///     wait for a response. If the ioctl times out before a response is
///     received, then the ioctl is canceled and considered failed.
///
///     In this function, the synchronous ioctl is actually implemented as an
///     asynchronous ioctl, and then the function waits on the object with a
///     user specified timeout. If the ioctl times out before its completed,
///     then the ioctl is canceled and the funciton returns TERA_ERR_TIMEOUT.
///
///     If this object is not initialized or if the device is not ready for
///     I/O operations, then this function returns TERA_ERR_INVALID_STATE.
///
///     This function blocks until the ioctl has completed. In Linux the
///     ioctl does not support a timeout mechanism, so it's up to the driver
///     to implement this. Some IOCTL calls may take longer to complete because
///     they send commands to the client and wait for a response versus others
///     that query information within the driver itself. It's up to the driver
///     to implement different timeouts based on the specific transaction type.
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_INVALID_STATE,
///     TERA_ERR_TIMEOUT,
///     TERA_ERR_NOT_SUPPORTED,
///     TERA_ERR_FAILURE
///
INT32 cTERA_DEV::synch_ioctl(
    UINT32 code,       ///< Ioctl code
    void  *buf,        ///< Buffer
    UINT32 buf_len,    ///< Buffer length - unused in Linux port
    UINT32 timeout     ///< Timeout (ms)  - unused in Linux port
    )
{
    INT32 ret = TERA_SUCCESS;
    int   rv;

    // Check if this object is initialized first
    if (!m_initialized)
    {
        mTERA_LOG_ERR("not m_initialized");

        ret = TERA_ERR_INVALID_STATE;
    }
    else
    {
        //
        // Linux does not support synchronous ioctls with timeouts, so we rely on
        // the driver to implement a timeout mechanism that prevents this call
        // from blocking forever
        //
        rv = ioctl(m_dev_fd, code, buf);
        if (rv < 0)
        {
            mTERA_LOG_ERR("ioctl 0x%08X failed (rv=%d)", code, rv);
            ret = TERA_ERR_FAILURE;
        }
    }
    return ret;
} // cTERA_DEV::synch_ioctl()
