////////////////////////////////////////////////////////////////////////////
//   Copyright 2013  Teradici Corporation
//
//   This program is free software; you can redistribute it and/or modify it
//   under the terms and conditions of the GNU General Public License,
//   version 2, as published by the Free Software Foundation.
//
////////////////////////////////////////////////////////////////////////////
/// \file
/// \brief
///     PCoIP Host Software GUI Class
///
///     This file defines the PCoIP Host Software GUI class.
///
////////////////////////////////////////////////////////////////////////////

#pragma once

// ---- Include Files -------------------------------------------------------

#include <wx/frame.h>
#include <wx/timer.h>
#include <wx/window.h>

#include "tera_pcoip_agent_gui_tray.h"
#include "tera_pcoip_agent_gui_prop.h"
#include "tera_pcoip_link.h"
#include "tera_netif.h"
#include "tera_cfg.h"

#include <X11/Xlibint.h>
#include <X11/extensions/Xfixes.h>

#include <gdk/gdk.h>
#include <gdk/gdkx.h>

// ---- Constants and Types -------------------------------------------------

#if __WXDEBUG__
#define HOST_SW_VERSION_STRING  "4.1.6"
#else
#define HOST_SW_VERSION_STRING  "4.1.6"
#endif

// ---- User Defines -------------------------------------------------------

// --------------------------------------------------------------------------
//  eTERA_AGENT_GUI_STATE
///     States of the agent GUI
//
//  Description:
///     This enumeration defines the states of the GUI. The heartbeat process
///     executes periodically and performs various processing based on the
///     current state of the GUI.
///
typedef enum
{
    TERA_AGENT_GUI_DISCONNECTED_STATE = 0,
    TERA_AGENT_GUI_SESS_START_STATE,
    TERA_AGENT_GUI_READ_CLIENT_TOPO_STATE,
    TERA_AGENT_GUI_READ_HOST_TOPO_STATE,
    TERA_AGENT_GUI_PUSH_TOPO_STATE,
    TERA_AGENT_GUI_SEND_TOPO_TO_CLIENT_STATE,
    TERA_AGENT_GUI_CHECK_TOPO_CHANGE_STATE,
    TERA_AGENT_GUI_STATE_MAX
} eTERA_AGENT_GUI_STATE;

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI
///     PCoIP Host Software GUI
//
//  Description:
///     This class defines the main GUI portion of the PCoIP Host Software.
///     The main dialog is invisible and is used only for collecting
///     messages. All GUI components are members of this class.
///
class cPCOIP_AGENT_GUI : public wxFrame
{

public:

    // --- Public Functions -------------------------------------------------

    cPCOIP_AGENT_GUI(
        wxWindow       *parent,
        const wxString& title);

    ~cPCOIP_AGENT_GUI(void);

    int display_topology_send(void);

    int display_topology_ok(void);

    void host_mon_info_invalidate(void);

    void end_session(void);

    INT32 get_host_cursor_pos_abs(
        UINT16  *x_pos,
        UINT16  *y_pos);

    INT32 get_host_cursor_pos_pix(
        UINT16  *x_pos,
        UINT16  *y_pos);

    void lcur_lkbd_settings_send(
        int mon_info_ok);

    INT lcur_lkbd_enable_get(void);

    INT lcur_fade_get(void);

    void prop_sheet_show(void);

    void prop_sheet_close(void);

    int prop_sheet_is_shown(void);

    cTERA_PCOIP_LINK& pcoip_link(void);

    cTERA_NETIF& host_netif(void);

    cTERA_CFG& host_cfg(void);

    void get_client_topo(
        sTERA_PCOIP_LINK_CACHE_DISP_TOPO *disp_topo);

    void force_local_cursor_update(void);

    void hide_host_cursor(void);

    void show_host_cursor(void);

    void set_twinview_mode(
        INT enable);

    void set_lcur_lkbd_mode(
        INT enable);

    void set_disc_exit_mode(
        INT enable);

    void set_remote_cur_sync_settings(
        INT  enable,
        long period);

    void set_topo_sent_status(
        INT enable);

    int is_host_cur_hiding_supported(void);

    int is_user_logged_in_locally(void);

    // --- Public Variables -------------------------------------------------
    cPCOIP_AGENT_GUI_PROP m_prop_sheet;             ///< Property sheet
    INT                   m_disc_exit_popup_enabled;///< Flag that indicates whether the disconnect & exit
                                                    ///< options are shown on the agent right-click menu

protected:

    typedef struct sSCREEN_CFG_STRUCT
    {
        struct
        {
            INT32 x;
            INT32 y;
        } size;
        struct
        {
            INT32 x;
            INT32 y;
        } pos;
    } sSCREEN_CFG;

    // --- Protected Functions ----------------------------------------------

    // Event handlers
    void evt_init_dialog(
        wxInitDialogEvent& evt);

    void evt_close(
        wxCloseEvent& evt);

    void evt_end_session(
        wxCloseEvent& evt);

    void evt_timer(
        wxTimerEvent& evt);

    virtual void OnExit(
        wxCommandEvent& event);

    virtual void OnFatalException(void);

    // Local functions
    void link_state_change_proc(
        UINT8 connected);

    void heartbeat(void);

    void send_kbd_repeat_settings(void);

    bool screen_topology_changed(void);

    UINT32 get_num_screens(void);

    void screen_topology_init(void);

    INT32 client_cache_topology_get(void);

    void connect_proc(void);

    static GdkFilterReturn xfixes_events_cback(
        GdkXEvent *xevent,
        GdkEvent *event,
        gpointer data);

    void process_cursor_shape_change_event (void);

    void process_cursor(void);

    void update_client_cur_pos(void);

    bool init_screen_lock_utility(void);

    // --- Protected Variables ----------------------------------------------

    cPCOIP_AGENT_GUI_TRAY m_tray;                 ///< System tray icon
    wxTimer               m_heartbeat_timer;      ///< Heartbeat timer
    cTERA_PCOIP_LINK      m_pcoip_link;           ///< PCoIP link object
    cTERA_NETIF           m_host_netif;           ///< Host network interfaces
    cTERA_CFG             m_host_cfg;             ///< Configuration data
    INT                   m_sess_connected;       ///< Active session flag
    eTERA_AGENT_GUI_STATE m_agent_state;          ///< state of the agent GUI
    INT                   m_read_client_topo_retry;///< client topology read retry counter
    INT                   m_client_send_topo_retry; ///< limits the # of times the topology is sent to the client device
    INT                   m_twinview_active;      ///< Flag that indicates whether TwinView mode is active
    UINT32                m_screen_count;         ///< Number of active screens
    sSCREEN_CFG           m_screen_cfg[TERA_PCOIP_HCC_APDU_MAX_NUM_DISP]; ///< Screen configuration settings
    INT32                 m_primary_width;        ///< Width of first display
    INT32                 m_primary_height;       ///< Height of first display
    UINT32                m_last_feat_stat;       ///< status of the last received feature status APDU
    UINT32                m_last_feat_stat_cnt;   ///< count of the last received feature status APDU
    INT32                 m_client_cache_status;  ///< status of last attempt to read the client cached topology
    INT32                 m_lcur_lkbd_en;         ///< status of local cursor and keyboard feature
    INT                   m_host_cur_hidden;      ///< Set when the host cursor is hidden
    INT                   m_host_cur_hide_supported; ///< Set when the system supports hiding the host cursor
    unsigned long         m_last_cursor_serial;   ///< serial number of the last cursor (used to check if the cursor changed)
    Display              *m_xdisplay;             ///< pointer to X Window Display structure
    INT                   m_topo_sent_ok;         ///< status of last attempt to send topology to the client
                                                  ///< the local cursor feature should not be enabled when the
                                                  ///< topology isn't set properly
    UINT32                m_dev_type;             ///< PCoIP host card device type
    wxWindow             *m_parent;
    sTERA_PCOIP_LINK_CACHE_DISP_TOPO m_disp_topo; ///< Client topology settings
    INT                   m_rcs_enabled;          ///< Flag that indicates whether the host SW sends
                                                  ///< remote cursor position changes to the client
    long                  m_rcs_period;           ///< period (in number of milliseconds) the RCS process runs
    wxString              m_lock_utility;         ///< contains string executed to lock the workstation
    INT                   m_icon_create_delay;    ///< Counter used to delay the creation of the tray icon

    DECLARE_EVENT_TABLE()

    // --- Private Variables ------------------------------------------------

}; // cPCOIP_AGENT_GUI

// ---- Variable Externs ----------------------------------------------------

//extern cPCOIP_AGENT_GUI *g_gui;

// ---- Function Prototypes -------------------------------------------------

