////////////////////////////////////////////////////////////////////////////
//   Copyright 2013  Teradici Corporation
//
//   This program is free software; you can redistribute it and/or modify it
//   under the terms and conditions of the GNU General Public License,
//   version 2, as published by the Free Software Foundation.
//
////////////////////////////////////////////////////////////////////////////
/// \file
/// \brief
///     PCoIP Link Class Imaging Functions Implementation
///
///     TODO UPDATE
///     This file implements link class environment feature. All environment
///     feature related operations are dealt with in this file. This feature
///     offers:
///         - sending host monitor information
///         - request client topology settings
///         - get/set user config settings (image quality)
///
////////////////////////////////////////////////////////////////////////////

// ---- Include Files -------------------------------------------------------

#include "tera_types.h"
#include "tera_log.h"
#include "tera_errors.h"
#include "tera_mem.h"
#include "tera_pcoip_link.h"
#include "tera_pcoip_hcc_ioctl.h"
#include "tera_pcoip_hdc_ioctl.h"
#include "tera_dev.h"

// ---- Global Variables ----------------------------------------------------

// ---- Private Constants and Types -----------------------------------------

// The host SW supports a feature where the host can update the position
// of the client cursor. Periodically the GUI polls the current host cursor
// position and checks if the cursor moved. If the cursor moved the GUI
// sends the new cursor position to the driver. The driver checks if the host
// cursor location equals one of the last X received client cursor reports. If
// the cursor locations differ the host cursor position is sent to the client.
//
// The following constant is used to compute a value used by the function that
// compares the current position of the host cursor to a client cursor point.
// The function compares the 2 points and declares whether the 2 points are the
// same. Since the 2 points are generated by 2 different entities with
// differing precision the 2 points may be slightly different. This constant
// determines how different the 2 points can be but still be considered equal.
#define MOUSE_POS_DIFF_EPSILON_IN_PIXEL ( 3 )   // 3 pixels is the threshold

// ---- Local Variables -----------------------------------------------------

// ---- Local Function Prototypes -------------------------------------------

// --------------------------------------------------------------------------
//  cTERA_PCOIP_LINK::disp_topology_get
///     Get the display topology from the firmware
//
//  Description:
///     This function gets the display topology from the firmware. The topology
///     in the firmware is only valid during a PCoIP session.
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_INVALID_ARG,
///     TERA_ERR_TIMEOUT,
///     TERA_ERR_FAILURE
///
INT32 cTERA_PCOIP_LINK::disp_topology_get(
    sTERA_PCOIP_HCC_APDU_IMG_DISP_TOPO *disp_topo
    )
{
    INT32 ret;

    mTERA_LOG_DBG("start");

    if (disp_topo == NULL)
    {
        mTERA_LOG_ERR("invalid arguments");
        ret = TERA_ERR_INVALID_ARG;
    }
    else
    {
        // Send a synch ioctl to the device
        ret = m_hcc_dev.synch_ioctl(TERA_PCOIP_HCC_IOCTL_IMG_DISP_TOPO_GET,
                                    disp_topo,
                                    sizeof(sTERA_PCOIP_HCC_APDU_IMG_DISP_TOPO),
                                    1000);
        if (ret != TERA_SUCCESS)
        {
            mTERA_LOG_ERR("HCC_IOCTL_IMG_DISP_TOPO_GET ioctl failed (ret=%d)",
                          ret);
        }
        else
        {
            // Dump the data inside the APDU
            int i;
            for (i = 0; i < TERA_PCOIP_HCC_APDU_MAX_NUM_DISP; i++)
            {
                mTERA_LOG_DBG("disp%u: pos_x=%d, pos_y=%d, width=%u, height=%u, rotation=%u",
                              i, disp_topo->disp[i].pos_x,
                              disp_topo->disp[i].pos_y,
                              disp_topo->disp[i].width,
                              disp_topo->disp[i].height,
                              disp_topo->disp[i].rotation);
            }
        }
    }

    mTERA_LOG_DBG("end");

    return ret;
} // cTERA_PCOIP_LINK::disp_topology_get()

// --------------------------------------------------------------------------
//  cTERA_PCOIP_LINK::async_disp_topology_get
///     Get the last received asynchronous display topology message
//
//  Description:
///     This function gets a copy of the last asynchronous display topology
///     message received by the driver from the firmware. At the start of a
///     session the client sends this message if its running firmware release
///     3.2.x or higher.
///
///     The soft client will send this message to the host SW when it wants to
///     change the host display topology.
///
//  Notes:
///     At one point we planned to push these settings into the host OS, but
///     another mechanism was created to do this because additional information
///     (EDID) was needed. This is the cached topology message. Both the ZC and
///     the SC must send the cached topology message because it contains needed
///     EDID information used by the host SW to match the client monitors with
///     the host OS display adapters.
///
///     The id parameter returns a value that is incremented by the driver each
///     time a new async message is received. This can be used by the caller to
///     quickly determine whether a new message has been received.
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_INVALID_ARG,
///     TERA_ERR_TIMEOUT,
///     TERA_ERR_FAILURE
///
INT32 cTERA_PCOIP_LINK::async_disp_topology_get(
    sTERA_PCOIP_LINK_DISP_TOPO *disp_topo,
    UINT32                     *id
    )
{
    sTERA_PCOIP_HCC_APDU_IMG_ASYNC_DISP_TOPO async_disp_topo;
    INT32                                    ret;
    INT                                      i;
    UINT32                                   num_disp=0;

//    mTERA_LOG_DBG("start");

    if ((disp_topo == NULL) || (id == NULL))
    {
        mTERA_LOG_ERR("invalid arguments");
        ret = TERA_ERR_INVALID_ARG;
    }
    else
    {
        // Send a synch ioctl to the device
        ret = m_hcc_dev.synch_ioctl(TERA_PCOIP_HCC_IOCTL_IMG_ASYNC_DISP_TOPO_GET,
                                    &async_disp_topo,
                                    sizeof(sTERA_PCOIP_HCC_APDU_IMG_ASYNC_DISP_TOPO),
                                    1000);
        if (ret != TERA_SUCCESS)
        {
            mTERA_LOG_ERR("HCC_IOCTL_IMG_ASYNC_DISP_TOPO_GET ioctl failed (ret=%d)",
                          ret);
        }
        else
        {
            // Copy the topology data to the callers structure.
            mTERA_MEM_ZERO(disp_topo, sizeof(sTERA_PCOIP_LINK_DISP_TOPO));
            for (i=0; i<TERA_PCOIP_HCC_APDU_MAX_NUM_DISP; i++)
            {
                if ((async_disp_topo.apdu.disp[i].width) ||
                    (async_disp_topo.apdu.disp[i].height))
                {
                    disp_topo->disp[i].x_origin = async_disp_topo.apdu.disp[i].pos_x;
                    disp_topo->disp[i].y_origin = async_disp_topo.apdu.disp[i].pos_y;
                    disp_topo->disp[i].width    = async_disp_topo.apdu.disp[i].width;
                    disp_topo->disp[i].height   = async_disp_topo.apdu.disp[i].height;
                    disp_topo->disp[i].rotation =
                        (async_disp_topo.apdu.disp[i].rotation == TERA_PCOIP_HCC_APDU_IMG_ROTATION_0) ?
                        TERA_PCOIP_LINK_DISP_ROTATION_0 :
                        (async_disp_topo.apdu.disp[i].rotation == TERA_PCOIP_HCC_APDU_IMG_ROTATION_90) ?
                        TERA_PCOIP_LINK_DISP_ROTATION_90 :
                        (async_disp_topo.apdu.disp[i].rotation == TERA_PCOIP_HCC_APDU_IMG_ROTATION_180) ?
                        TERA_PCOIP_LINK_DISP_ROTATION_180 : TERA_PCOIP_LINK_DISP_ROTATION_270;

                    num_disp++;
                }
            }
            disp_topo->num_disp = num_disp;
            *id = async_disp_topo.msg_id;
        }
    }

//    mTERA_LOG_DBG("end");

    return ret;
} // cTERA_PCOIP_LINK::async_disp_topology_get()

// --------------------------------------------------------------------------
//  cTERA_PCOIP_LINK::cache_disp_topology_get
///     Get the client cache display topology settings
//
//  Description:
///     This function retrieves the cached display topology settings from the
///     client. These settings are stored in FLASH on the zero client.
///
///     Note: the message returned from the client includes a field that
///     specifies which display is the primary. This field is ignored by the
///     host SW, which instead makes the primary display the display with its
///     origin equal to (0,0).
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_INVALID_ARG,    (bad input parameter)
///     TERA_ERR_TIMEOUT,
///     TERA_ERR_FAILURE,
///     TERA_ERR_INVALID_STATE,  (bad refresh rate)
///     TERA_ERR_NOT_SUPPORTED,  (FW does not support get cached topo APDU)
///     TERA_ERR_EMPTY           (soft client OR client topo disabled)
///
INT32 cTERA_PCOIP_LINK::cache_disp_topology_get(
    sTERA_PCOIP_LINK_CACHE_DISP_TOPO *disp_topo
    )
{
    sTERA_PCOIP_HCC_APDU_IMG_CACHE_DISP_TOPO cache_disp_topo;
    INT32                                    ret;
    INT                                      i;
    UINT32                                   num_disp=0;

    if (disp_topo == NULL)
    {
        mTERA_LOG_ERR("invalid argument");
        ret = TERA_ERR_INVALID_ARG;
    }
    else
    {
        // initialize the structure returned to the caller
        mTERA_MEM_ZERO(disp_topo, sizeof(sTERA_PCOIP_LINK_CACHE_DISP_TOPO));

        if (TERA_ERR_NOT_SUPPORTED == cached_topo_get_avail())
        {
            mTERA_LOG_DBG("FW does not support get cached topology");
            ret = TERA_ERR_NOT_SUPPORTED;
        }
        else
        {
            // fill in the required fields in the get APDU
            cache_disp_topo.apdu_type = TERA_PCOIP_HCC_APDU_IMG_TYPE_CACHE_DISP_TOPO_GET;
            cache_disp_topo.pri_num = 0;

            // Send a synch ioctl to the device
            ret = m_hcc_dev.synch_ioctl(TERA_PCOIP_HCC_IOCTL_IMG_CACHE_DISP_TOPO_GET,
                                        &cache_disp_topo,
                                        sizeof(sTERA_PCOIP_HCC_APDU_IMG_CACHE_DISP_TOPO),
                                        2000);
            if (ret != TERA_SUCCESS)
            {
                mTERA_LOG_ERR("HCC_IOCTL_IMG_CACHE_DISP_TOPO_GET ioctl failed (ret=%d)",
                              ret);
            }
            else
            {
                // Copy the returned topology data into the callers structure.
                for (i=0; i<TERA_PCOIP_HCC_APDU_MAX_NUM_DISP; i++)
                {
                    // if this entry is not empty
                    if ((cache_disp_topo.topo.topology[i].width) ||
                        (cache_disp_topo.topo.topology[i].height))
                    {
                        num_disp++;

                        // verify the refresh rate is valid
                        if ((cache_disp_topo.topo.refresh_rate[i] != 56) &&
                            (cache_disp_topo.topo.refresh_rate[i] != 59) &&
                            (cache_disp_topo.topo.refresh_rate[i] != 60) &&
                            (cache_disp_topo.topo.refresh_rate[i] != 65) &&
                            (cache_disp_topo.topo.refresh_rate[i] != 67) &&
                            (cache_disp_topo.topo.refresh_rate[i] != 70) &&
                            (cache_disp_topo.topo.refresh_rate[i] != 72) &&
                            (cache_disp_topo.topo.refresh_rate[i] != 73) &&
                            (cache_disp_topo.topo.refresh_rate[i] != 75) &&
                            (cache_disp_topo.topo.refresh_rate[i] != 82) &&
                            (cache_disp_topo.topo.refresh_rate[i] != 87) &&
                            (cache_disp_topo.topo.refresh_rate[i] != 88))
                        {
                            mTERA_LOG_ERR("disp%i: invalid refresh rate = %u",
                                          i, cache_disp_topo.topo.refresh_rate[i]);
                            ret = TERA_ERR_INVALID_STATE;
                        }
                    }
                    //
                    // This statement is true for 2 reasons:
                    // - the client does not support the feature, which happens
                    //   when a HH is talking to a SC
                    // - the ZC display topology settings are disabled
                    //
                    else if (cache_disp_topo.topo.refresh_rate[i] ==
                             TERA_PCOIP_HCC_HOST_TOPOLOGY_RATE_NO_CACHING)
                    {
                        num_disp++;
                        cache_disp_topo.topo.refresh_rate[i] = 0;
                        ret = TERA_ERR_EMPTY;
                    }
                    // This statement is true if imaging is not yet active
                    else if (cache_disp_topo.topo.refresh_rate[i] ==
                             TERA_PCOIP_HCC_HOST_TOPOLOGY_RATE_NOT_ACTIVE)
                    {
                        num_disp++;
                        cache_disp_topo.topo.refresh_rate[i] = 0;
                        ret = TERA_ERR_IMG_NOT_ACTIVE;
                    }
                    disp_topo->disp[i].x_origin = cache_disp_topo.topo.topology[i].pos_x;
                    disp_topo->disp[i].y_origin = cache_disp_topo.topo.topology[i].pos_y;
                    disp_topo->disp[i].width    = cache_disp_topo.topo.topology[i].width;
                    disp_topo->disp[i].height   = cache_disp_topo.topo.topology[i].height;
                    disp_topo->disp[i].rotation =
                        (cache_disp_topo.topo.topology[i].rotation == TERA_PCOIP_HCC_APDU_IMG_ROTATION_0) ?
                        TERA_PCOIP_LINK_DISP_ROTATION_0 :
                        (cache_disp_topo.topo.topology[i].rotation == TERA_PCOIP_HCC_APDU_IMG_ROTATION_90) ?
                        TERA_PCOIP_LINK_DISP_ROTATION_90 :
                        (cache_disp_topo.topo.topology[i].rotation == TERA_PCOIP_HCC_APDU_IMG_ROTATION_180) ?
                        TERA_PCOIP_LINK_DISP_ROTATION_180 : TERA_PCOIP_LINK_DISP_ROTATION_270;
                    disp_topo->disp[i].refresh_rate = cache_disp_topo.topo.refresh_rate[i];

                    //
                    // Note the length of the strings are different. The length in
                    // the APDU is longer than necessary. The following code copies
                    // the required number of characters.
                    //
                    mTERA_MEM_COPY(disp_topo->iden[i].vid,
                                   (char *)cache_disp_topo.iden[i].vid,
                                   TERA_DISP_IDEN_VID_LEN);
                    mTERA_MEM_COPY(disp_topo->iden[i].pid,
                                   (char *)cache_disp_topo.iden[i].pid,
                                   TERA_DISP_IDEN_PID_LEN);
                    mTERA_MEM_COPY(disp_topo->iden[i].serial,
                                   (char *)cache_disp_topo.iden[i].serial,
                                   TERA_DISP_IDEN_SERIAL_LEN);
                    mTERA_MEM_COPY(disp_topo->iden[i].name,
                                   (char *)cache_disp_topo.iden[i].name,
                                   TERA_DISP_IDEN_NAME_LEN);
                }
                disp_topo->num_disp = num_disp;
                disp_topo->primary = cache_disp_topo.topo.primary;
            }
        }
    }

    if (ret == TERA_ERR_EMPTY)
    {
        mTERA_LOG_DBG("FW returned TERA_PCOIP_HCC_HOST_TOPOLOGY_RATE_NO_CACHING");
    }
    else if (ret == TERA_ERR_IMG_NOT_ACTIVE)
    {
        mTERA_LOG_DBG("FW returned TERA_PCOIP_HCC_HOST_TOPOLOGY_RATE_NOT_ACTIVE");
    }

    // Dump the data returned to the caller
    mTERA_LOG_DBG("primary=%d",
                  disp_topo->primary);
    for (i = 0; i < TERA_PCOIP_HCC_APDU_MAX_NUM_DISP; i++)
    {
        mTERA_LOG_DBG("disp%u: %u x %u @ (%d, %d) rot=%u rate=%u",
                      i,
                      disp_topo->disp[i].width,
                      disp_topo->disp[i].height,
                      disp_topo->disp[i].x_origin,
                      disp_topo->disp[i].y_origin,
                      disp_topo->disp[i].rotation,
                      disp_topo->disp[i].refresh_rate);
        mTERA_LOG_DBG("       name='%s', serial='%s', vid='%s', pid='%s'",
                      disp_topo->iden[i].name,
                      disp_topo->iden[i].serial,
                      disp_topo->iden[i].vid,
                      disp_topo->iden[i].pid);
    }

    mTERA_LOG_DBG("end, ret=%d, disp_topo->num_disp=%d", ret, disp_topo->num_disp);

    return ret;
} // cTERA_PCOIP_LINK::cache_disp_topology_get()

// --------------------------------------------------------------------------
//  cTERA_PCOIP_LINK::cache_disp_topology_set
///     Set the client cache display topology settings
//
//  Description:
///     This function sets the client cached display topology settings.
///     The new settings will be stored in FLASH on the zero client.
///
///     The caller passes in the swap_portrait_en flag which indicates
///     portrait swapping is enabled. When this is true the host SW should
///     consider the 90 and 270 degree display orientation settings reversed.
///     This is done to deal with GPUs that inverted the values of these
///     settings. If the settings are not swapped the local cursor and
///     Windows cursor do not synchronized when the display is rotated 90 or
///     270 degrees.
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_INVALID_ARG,
///     TERA_ERR_TIMEOUT,
///     TERA_ERR_FAILURE,
///     TERA_ERR_INVALID_STATE
///
INT32 cTERA_PCOIP_LINK::cache_disp_topology_set(
    sTERA_DISP_TOPO            *host_disp_topo,
    sTERA_DISP_EDID            *host_disp_edid,
    UINT32                      swap_portrait_en
    )
{
    sTERA_PCOIP_HCC_APDU_IMG_CACHE_DISP_TOPO cache_disp_topo;
    INT32                                    ret;
    INT                                      i;
    INT                                      primary_set = 0;

    mTERA_LOG_DBG("start");

    if ((host_disp_topo == NULL) || (host_disp_edid == NULL))
    {
        mTERA_LOG_ERR("invalid argument");
        ret = TERA_ERR_INVALID_ARG;
    }
    else
    {
        // fill in the required fields in the set APDU
        mTERA_MEM_ZERO(&cache_disp_topo, sizeof(cache_disp_topo));
        cache_disp_topo.apdu_type = TERA_PCOIP_HCC_APDU_IMG_TYPE_CACHE_DISP_TOPO_SET;
        cache_disp_topo.pri_num = 0;

        for (i=0; i<TERA_PCOIP_HCC_APDU_MAX_NUM_DISP; i++)
        {
            cache_disp_topo.topo.topology[i].pos_x  = host_disp_topo->disp[i].x_origin;
            cache_disp_topo.topo.topology[i].pos_y  = host_disp_topo->disp[i].y_origin;
            cache_disp_topo.topo.topology[i].height = host_disp_topo->disp[i].height;
            cache_disp_topo.topo.topology[i].width  = host_disp_topo->disp[i].width;
            if (0 == swap_portrait_en)
            {
                cache_disp_topo.topo.topology[i].rotation =
                        (host_disp_topo->disp[i].rotation == TERA_DISP_ROTATION_0) ?
                        TERA_PCOIP_HCC_APDU_IMG_ROTATION_0 :
                        (host_disp_topo->disp[i].rotation == TERA_DISP_ROTATION_90) ?
                        TERA_PCOIP_HCC_APDU_IMG_ROTATION_90 :
                        (host_disp_topo->disp[i].rotation == TERA_DISP_ROTATION_180) ?
                        TERA_PCOIP_HCC_APDU_IMG_ROTATION_180 : TERA_PCOIP_HCC_APDU_IMG_ROTATION_270;
            }
            else
            {
                cache_disp_topo.topo.topology[i].rotation =
                        (host_disp_topo->disp[i].rotation == TERA_DISP_ROTATION_0) ?
                        TERA_PCOIP_HCC_APDU_IMG_ROTATION_0 :
                        (host_disp_topo->disp[i].rotation == TERA_DISP_ROTATION_90) ?
                        TERA_PCOIP_HCC_APDU_IMG_ROTATION_270 :
                        (host_disp_topo->disp[i].rotation == TERA_DISP_ROTATION_180) ?
                        TERA_PCOIP_HCC_APDU_IMG_ROTATION_180 : TERA_PCOIP_HCC_APDU_IMG_ROTATION_90;
                if ((TERA_PCOIP_HCC_APDU_IMG_ROTATION_90 == cache_disp_topo.topo.topology[i].rotation) ||
                    (TERA_PCOIP_HCC_APDU_IMG_ROTATION_270 == cache_disp_topo.topo.topology[i].rotation))
                {
                    mTERA_LOG_DBG("swapped written cached client orientation data for disp%d", i);
                }
            }
            cache_disp_topo.topo.refresh_rate[i] = host_disp_topo->disp[i].refresh_rate;

            strcpy((char *)cache_disp_topo.iden[i].vid,
                   host_disp_edid->edid[i].vid);
            strcpy((char *)cache_disp_topo.iden[i].pid,
                   host_disp_edid->edid[i].pid);
            strcpy((char *)cache_disp_topo.iden[i].serial,
                   host_disp_edid->edid[i].serial);
            strcpy((char *)cache_disp_topo.iden[i].name,
                   host_disp_edid->edid[i].name);

            //
            // determine which display is the primary, the primary is the monitor
            // with its origin = (0,0) and a non-blank geometry
            //
            if ((host_disp_topo->disp[i].x_origin == 0) &&
                (host_disp_topo->disp[i].y_origin == 0) &&
                (host_disp_topo->disp[i].height != 0) &&
                (host_disp_topo->disp[i].width != 0))
            {
                cache_disp_topo.topo.primary = i;
                primary_set = 1;
            }
        }

        if (!primary_set)
        {
            mTERA_LOG_ERR("No primary configured!!!");
        }

        // Send a synch ioctl to the device
        ret = m_hcc_dev.synch_ioctl(TERA_PCOIP_HCC_IOCTL_IMG_CACHE_DISP_TOPO_SET,
                                    &cache_disp_topo,
                                    sizeof(sTERA_PCOIP_HCC_APDU_IMG_CACHE_DISP_TOPO),
                                    1000);
        if (ret != TERA_SUCCESS)
        {
            mTERA_LOG_ERR("HCC_IOCTL_IMG_CACHE_DISP_TOPO_SET ioctl failed (ret=%d)",
                          ret);
        }
    }

    mTERA_LOG_DBG("end");

    return ret;
} // cTERA_PCOIP_LINK::cache_disp_topology_set()

// --------------------------------------------------------------------------
//  cTERA_PCOIP_LINK::disp_topology_send
///     Send the display topology to the firmware
//
//  Description:
///     This function sends the display topology to the firmware. The APDU type
///     and PRI number does not need to be filled out by the caller.
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_INVALID_ARG,
///     TERA_ERR_TIMEOUT,
///     TERA_ERR_FAILURE
///
INT32 cTERA_PCOIP_LINK::disp_topology_send(
        sTERA_PCOIP_LINK_DISP_TOPO *disp_topo,
        INT32                       primary_width,
        INT32                       primary_height,
        INT                         twinview_active
    )
{
    INT32                                      ret;
    sTERA_PCOIP_HCC_APDU_HMI_VAR_SESS_LINK     sess_link;
    sTERA_PCOIP_HCC_APDU_IMG_DISP_TOPO         hcc_buf;
    sTERA_PCOIP_HDC_LOCAL_CURSOR_DISPLAY_TABLE hdc_buf;
    UINT32                                     i;

    mTERA_LOG_DBG("start");

    if ((disp_topo == NULL) ||
        (disp_topo->num_disp > TERA_PCOIP_LINK_MAX_NUM_DISP) ||
        (primary_width == 0) ||
        (primary_height == 0))
    {
        mTERA_LOG_ERR("invalid argument: %d X %d", primary_width, primary_height);

        ret = TERA_ERR_INVALID_ARG;
    }
    else
    {
        // Create the message sent to the HCC driver
        mTERA_MEM_ZERO(&hcc_buf, sizeof(hcc_buf));
        hcc_buf.apdu_type = TERA_PCOIP_HCC_APDU_IMG_TYPE_DISP_TOPO;
        hcc_buf.pri_num = 0;

        for (i = 0; i < TERA_PCOIP_HCC_APDU_MAX_NUM_DISP; i++)
        {
            hcc_buf.disp[i].pos_x = disp_topo->disp[i].x_origin;
            hcc_buf.disp[i].pos_y = disp_topo->disp[i].y_origin;
            hcc_buf.disp[i].width = disp_topo->disp[i].width;
            hcc_buf.disp[i].height = disp_topo->disp[i].height;
            hcc_buf.disp[i].rotation = (eTERA_PCOIP_HCC_APDU_IMG_ROTATION)disp_topo->disp[i].rotation; //TODO fix

            mTERA_LOG_DBG("disp%u: %d x %d @ (%d, %d) rot=%u",
                          i, hcc_buf.disp[i].width, hcc_buf.disp[i].height,
                          hcc_buf.disp[i].pos_x, hcc_buf.disp[i].pos_y,
                          hcc_buf.disp[i].rotation);
        }

        // Create the message sent to the HDC driver
        mTERA_MEM_ZERO(&hdc_buf, sizeof(hdc_buf));
        hdc_buf.primary_width = primary_width;
        hdc_buf.primary_height = primary_height;
        hdc_buf.twinview_active = twinview_active;

        for (i=0; i<TERA_PCOIP_HCC_APDU_MAX_NUM_DISP; i++)
        {
            if ((hcc_buf.disp[i].height) && (hcc_buf.disp[i].width))
            {
                hdc_buf.num_active_displays++;
                hdc_buf.displays[i].active    = 1;
                hdc_buf.displays[i].min_pix_x = hcc_buf.disp[i].pos_x;
                hdc_buf.displays[i].min_pix_y = hcc_buf.disp[i].pos_y;
                hdc_buf.displays[i].max_pix_x = hcc_buf.disp[i].pos_x + hcc_buf.disp[i].width - 1;
                hdc_buf.displays[i].max_pix_y = hcc_buf.disp[i].pos_y + hcc_buf.disp[i].height - 1;
                hdc_buf.displays[i].width     = hcc_buf.disp[i].width;
                hdc_buf.displays[i].height    = hcc_buf.disp[i].height;
                hdc_buf.displays[i].rotation  = hcc_buf.disp[i].rotation;
                if (hdc_buf.displays[i].min_pix_x < hdc_buf.min_x)
                {
                    hdc_buf.min_x = hdc_buf.displays[i].min_pix_x;
                }
                if (hdc_buf.displays[i].min_pix_y < hdc_buf.min_y)
                {
                    hdc_buf.min_y = hdc_buf.displays[i].min_pix_y;
                }
                if (hdc_buf.displays[i].max_pix_x > hdc_buf.max_x)
                {
                    hdc_buf.max_x = hdc_buf.displays[i].max_pix_x;
                }
                if (hdc_buf.displays[i].max_pix_y > hdc_buf.max_y)
                {
                    hdc_buf.max_y = hdc_buf.displays[i].max_pix_y;
                }
            }
        }

        if (hdc_buf.num_active_displays)
        {
            mTERA_LOG_DBG("twinview %d, primary %d x %d",
                          hdc_buf.twinview_active,
                          hdc_buf.primary_width,
                          hdc_buf.primary_height);

            hdc_buf.desktop_width  = abs(hdc_buf.min_x) + abs(hdc_buf.max_x) + 1;
            hdc_buf.desktop_height = abs(hdc_buf.min_y) + abs(hdc_buf.max_y) + 1;

            // calculate the X and Y tolerance values
            hdc_buf.desktop_x_tolerance = (65536 * MOUSE_POS_DIFF_EPSILON_IN_PIXEL)/hdc_buf.desktop_width;
            hdc_buf.desktop_y_tolerance = (65536 * MOUSE_POS_DIFF_EPSILON_IN_PIXEL)/hdc_buf.desktop_height;

            for (i=0; i<TERA_PCOIP_HCC_APDU_MAX_NUM_DISP; i++)
            {
                if ((hdc_buf.displays[i].height) && (hdc_buf.displays[i].width))
                {
                    hdc_buf.displays[i].min_abs_x = (hdc_buf.displays[i].min_pix_x * 65536) / hdc_buf.desktop_width;
                    hdc_buf.displays[i].min_abs_y = (hdc_buf.displays[i].min_pix_y * 65536) / hdc_buf.desktop_height;
                    hdc_buf.displays[i].max_abs_x = (((hdc_buf.displays[i].max_pix_x + 1) * 65536) / hdc_buf.desktop_width) - 1;
                    hdc_buf.displays[i].max_abs_y = (((hdc_buf.displays[i].max_pix_y + 1) * 65536) / hdc_buf.desktop_height) - 1;
                }
            }
        }

        // Only send if session is up
        if ((sess_link_get(&sess_link) != TERA_SUCCESS) ||
            !(sess_link.flags &
              TERA_PCOIP_HCC_APDU_HMI_VAR_SESS_LINK_FLAGS_CONNECTED))
        {
            //
            // We either failed to get the session link or there is no active
            // session, so we shouldn't send the display topology down
            //
            ret = TERA_ERR_FAILURE;
            mTERA_LOG_DBG("disp topology not sent (failed to get topology or no active session");
        }
        else
        {
            // Send a synch ioctl to the HCC driver, which forwards the message
            // to the host card.
            ret = m_hcc_dev.synch_ioctl(TERA_PCOIP_HCC_IOCTL_IMG_DISP_TOPO_SEND,
                                        &hcc_buf, sizeof(hcc_buf),
                                        1000);
            if (ret != TERA_SUCCESS)
            {
                mTERA_LOG_ERR("HCC_IOCTL_IMG_DISP_TOPO_SEND ioctl failed "
                              "(ret=%d)", ret);
            }
            else
            {
                mTERA_LOG_DBG("sent display topology to FW through HCC");

                // Send a synch ioctl to the HDC driver, which updates data in
                // the control block used by the driver while processing client
                // cursor data.
                ret = m_hcc_dev.synch_ioctl(TERA_PCOIP_HDC_IOCTL_KMP_DISP_TOPO_SET,
                                            &hdc_buf, sizeof(hdc_buf),
                                            1000);
                if (ret != TERA_SUCCESS)
                {
                    mTERA_LOG_ERR("HDC_IOCTL_KMP_DISP_TOPO_SET ioctl failed "
                                  "(ret=%d)", ret);
                }
                else
                {
                    mTERA_LOG_DBG("sent display topology to HDC");
                }
            }
        }
    } // end if

    mTERA_LOG_DBG("end");

    return ret;
} // cTERA_PCOIP_LINK::disp_topology_send()

// --------------------------------------------------------------------------
//  cTERA_PCOIP_LINK::img_user_cfg_req_get
///     Get the active imaging user config settings
//
//  Description:
///     This function retrieves the current imaging user config settings from
///     FW.
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_INVALID_ARG,    (bad input parameter)
///     TERA_ERR_TIMEOUT,
///     TERA_ERR_FAILURE,
///
INT32 cTERA_PCOIP_LINK::img_user_cfg_req_get(
    sTERA_PCOIP_LINK_IMG_USER_CFG *user_cfg
    )
{
    sTERA_PCOIP_HCC_APDU_IMG_USER_CFG_REQ_DATA  user_cfg_req_apdu;
    INT32                                       ret;

    if (user_cfg == NULL)
    {
        mTERA_LOG_ERR("invalid argument");
        ret = TERA_ERR_INVALID_ARG;
    }
    else if (TERA_ERR_NOT_SUPPORTED == img_user_ext_cfg_get_avail())
    {
        mTERA_LOG_DBG("HCC_IOCTL_IMG_USER_CFG_REQ_GET ioctl not supported");
        ret = TERA_ERR_NOT_SUPPORTED;
    }
    else
    {
        // initialize the structure returned to the caller
        mTERA_MEM_ZERO(user_cfg, sizeof(sTERA_PCOIP_LINK_IMG_USER_CFG));

        // fill in the required fields in the get APDU
        user_cfg_req_apdu.host_user_cfg_req.apdu_type = TERA_PCOIP_HCC_APDU_IMG_TYPE_HOST_USER_CFG_REQ;
        user_cfg_req_apdu.host_user_cfg_req.pri_num = 0;

        // Send a synch ioctl to the device
        ret = m_hcc_dev.synch_ioctl(TERA_PCOIP_HCC_IOCTL_IMG_USER_CFG_REQ_GET,
                                    &user_cfg_req_apdu,
                                    sizeof(sTERA_PCOIP_HCC_APDU_IMG_USER_CFG_REQ_DATA),
                                    1000);
        if (ret != TERA_SUCCESS)
        {
            mTERA_LOG_ERR("HCC_IOCTL_IMG_USER_CFG_REQ_GET ioctl failed (ret=%d)",
                          ret);
        }
        else
        {
            // Copy the returned data into the callers structure.
            user_cfg->image_qual_pref = user_cfg_req_apdu.user_cfg_settings.image_qual_pref;
            user_cfg->build_to_lossless_enable = user_cfg_req_apdu.user_cfg_settings.build_to_lossless_enable;

            mTERA_LOG_DBG("end, qual=%d, b2l=%d",
                          user_cfg->image_qual_pref,
                          user_cfg->build_to_lossless_enable);
        }
    }

    return ret;
} // cTERA_PCOIP_LINK::img_user_cfg_req_get()

// --------------------------------------------------------------------------
//  cTERA_PCOIP_LINK::img_user_cfg_set
///     Set the imaging user config settings
//
//  Description:
///     This function sets the imaging user config settings. The settings are
///     applied to the session.
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_INVALID_ARG,
///     TERA_ERR_TIMEOUT,
///     TERA_ERR_FAILURE,
///     TERA_ERR_INVALID_STATE
///
INT32 cTERA_PCOIP_LINK::img_user_cfg_set(
    UINT32      image_qual_pref
    )
{
    sTERA_PCOIP_HCC_APDU_IMG_USER_CFG_DATA user_cfg_set_apdu;
    INT32                                  ret;

    if (image_qual_pref > TERA_PCOIP_HCC_APDU_IMG_QUAL_PREF_MAX)
    {
        mTERA_LOG_ERR("invalid argument");
        ret = TERA_ERR_INVALID_ARG;
    }
    else if (TERA_ERR_NOT_SUPPORTED == img_user_ext_cfg_set_avail())
    {
        mTERA_LOG_DBG("HCC_IOCTL_IMG_USER_CFG_REQ_SET ioctl not supported");
        ret = TERA_ERR_NOT_SUPPORTED;
    }
    else
    {
        // fill in the required fields in the APDU
        user_cfg_set_apdu.host_user_cfg_req.apdu_type = TERA_PCOIP_HCC_APDU_IMG_TYPE_HOST_USER_CFG;
        user_cfg_set_apdu.host_user_cfg_req.pri_num = 0;
        user_cfg_set_apdu.user_cfg_settings.image_qual_pref = image_qual_pref;

        // Send a synch ioctl to the device
        ret = m_hcc_dev.synch_ioctl(TERA_PCOIP_HCC_IOCTL_IMG_USER_CFG_SET,
                                    &user_cfg_set_apdu,
                                    sizeof(sTERA_PCOIP_HCC_APDU_IMG_USER_CFG_DATA),
                                    1000);
        if (ret != TERA_SUCCESS)
        {
            mTERA_LOG_ERR("HCC_IOCTL_IMG_USER_CFG_SET ioctl failed (ret=%d)",
                          ret);
        }
    }

    return ret;
} // cTERA_PCOIP_LINK::img_user_cfg_set()

// --------------------------------------------------------------------------
//  cTERA_PCOIP_LINK::img_user_cfg_cache_set
///     Set the imaging user config cache settings
//
//  Description:
///     This function tells the client to permanently store the active imaging
///     user config settings.
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_INVALID_ARG,
///     TERA_ERR_TIMEOUT,
///     TERA_ERR_FAILURE,
///     TERA_ERR_INVALID_STATE
///
INT32 cTERA_PCOIP_LINK::img_user_cfg_cache_set(
        UINT32      image_qual_pref
        )
{
    sTERA_PCOIP_HCC_APDU_IMG_USER_CFG_CACHE_SET_DATA user_cfg_cache_set_apdu;
    INT32                                            ret;

    if (image_qual_pref > TERA_PCOIP_HCC_APDU_IMG_QUAL_PREF_MAX)
    {
        mTERA_LOG_ERR("invalid argument");
        ret = TERA_ERR_INVALID_ARG;
    }
    else if (TERA_ERR_NOT_SUPPORTED == img_user_ext_cfg_set_avail())
    {
        mTERA_LOG_DBG("HCC_IOCTL_IMG_USER_CFG_CACHE_SET ioctl not supported");
        ret = TERA_ERR_NOT_SUPPORTED;
    }
    else
    {
        // fill in the required fields in the APDU
        user_cfg_cache_set_apdu.host_user_cfg_req.apdu_type = TERA_PCOIP_HCC_APDU_IMG_TYPE_HOST_USER_CFG_CACHE_SET;
        user_cfg_cache_set_apdu.host_user_cfg_req.pri_num = 0;
        user_cfg_cache_set_apdu.image_qual_pref = image_qual_pref;

        // Send a synch ioctl to the device
        ret = m_hcc_dev.synch_ioctl(TERA_PCOIP_HCC_IOCTL_IMG_USER_CFG_CACHE_SET,
                                    &user_cfg_cache_set_apdu,
                                    sizeof(sTERA_PCOIP_HCC_APDU_IMG_USER_CFG_CACHE_SET_DATA),
                                    1000);
        if (ret != TERA_SUCCESS)
        {
            mTERA_LOG_ERR("HCC_IOCTL_IMG_USER_CFG_CACHE_SET ioctl failed (ret=%d)",
                          ret);
        }
    }

    return ret;
} // cTERA_PCOIP_LINK::img_user_cfg_cache_set()


